#!/usr/bin/env ruby
#
# Simple Text Editor
#
# The sample application of the FLTK documentation,
# done in Ruby-FLTK.
#
# @author Kent Dahl <MenThal@bigfoot.com>
#
# The original C++ version can be found at:
# - http://www.fltk.org/        a.k.a.
# - http://fltk.sourceforge.net/


require 'fltk'

class Editor

  WIDTH = 640
  HEIGHT = 480
  MENUBAR_YSIZE = 30

  def initialize
    # Obviously the first thing that we need is a window: 
    @window = Fltk::Window.new( WIDTH, HEIGHT, 'Text Editor')

    # Our text editor will need some variables 
    # to keep track of things: 
    @changed = false
    @filename = ""
    @search = nil

    # As of writing, the MenuItems haven't got Ruby bindings,
    # but we don't really need them.
    @menubar = Fltk::MenuBar.new( 0, 0, WIDTH, MENUBAR_YSIZE )

    # Shortcuts can be written as "^a" for Ctrl+A,
    # "#a" for Alt+A, "+a" for Shift+A, etc...
    # Or you can also do aritmetic like in regular FLTK:
    saveas_shortcut = Fltk::ALT + Fltk::SHIFT + ?s

    # When defining the menubar items, we can just call @menubar.add 
    # multiple times, but this way is more like the format in
    # the C++ example, and also more terse.
    #
    # Format is:
    # [ text, shortcut, callback, data, flags ]
    menuitems = [ 
      [ "&File",       "",   nil,       nil, Fltk::SUBMENU ],
      [ "&New",        "#n", :new_cb,   nil, 0],
      [ "&Open...",    "#o", :open_cb,  nil, Fltk::MENU_DIVIDER],
      [ "&Save...",    "#s", :save_cb,  nil, 0],
      [ "Save &As...", saveas_shortcut,	:saveas_cb,nil, Fltk::MENU_DIVIDER],
      [ "&Quit",       "#q", :quit_cb,  nil, 0],
      nil,
      [ "&Edit",   "",   nil,        nil, Fltk::SUBMENU ],
      [ "&Undo",   "#z", :undo_cb,   nil, Fltk::MENU_DIVIDER ],
      [ "Cu&t",    "#x", :cut_cb,    nil, 0],
      [ "&Copy",   "#c", :copy_cb,   nil, 0],
      [ "&Paste",  "#v", :paste_cb,  nil, 0],
      [ "&Delete", "",   :delete_cb, nil, 0],
      nil,
      [ "&Search",        "",   nil,          nil, Fltk::SUBMENU ],
      [ "&Find",          "#f", :find_cb,     nil, 0],
      [ "F&ind Again",    "#g", :find2_cb,    nil, 0],
      [ "&Replace",       "#r", :replace_cb,  nil, 0],
      [ "Re&place Again", "#t", :replace2_cb, nil, 0],
      nil
    ]
    
    # Call a helper method that expands the structure above into the @menubar
    setup_menubar( menuitems )
  
    # To keep things simple our text editor will use 
    # the Fl_Multiline_Input widget to edit the text: 
    @input = Fltk::MultilineInput.new(0, MENUBAR_YSIZE, WIDTH, HEIGHT-MENUBAR_YSIZE)

    # So that we can keep track of changes to the file,
    # we also want to add a "changed" callback: 
    @input.callback method(:changed_cb).to_proc
    @input.when( Fltk::WHEN_CHANGED )
  
    #Finally, we want to use a mono-spaced font like FL_COURIER: 
    @input.textfont(Fltk::COURIER)
  
    # The Replace Dialog
    @replace_dlg  = Fltk::Window.new(300,105,"Replace")
    @replace_find = Fltk::Input.new(70, 10, 200, 25, "Find:")

    @replace_with =   Fltk::Input.new(70, 40, 200, 25, "Replace:");
    @replace_all =    Fltk::Button.new(10, 70, 90, 25, "Replace All");
    @replace_next =   Fltk::ReturnButton.new(105, 70, 120, 25, "Replace Next");
    @replace_cancel = Fltk::Button.new(230, 70, 60, 25, "Cancel");
    @replace_all.callback    method( :replall_cb  ).to_proc
    @replace_next.callback   method( :replace2_cb ).to_proc
    @replace_cancel.callback method( :replcan_cb  ).to_proc

    @window.show

  end # initialize

  ##
  # This is a helper function to expand the two dimensional array
  # structure representing the menuitems into the @menubar
  #
  # The format is kept as similar as possible to the original
  # C++ exampleexample.
  # [ text, shortcut, callback, data, flags ]
  def setup_menubar( items )
    # menubar.add needs to have all levels prefixed ("File/Open"),
    # but we don't want to repeat our selves. 
    prefix_stack = Array.new
    
    items.each{ |e|
      if ! e
	# nil in the main array means that 
	# a sub-menu-group has ended
	prefix_stack.pop
	@menubar.add( "" )
      else
	# extract the data from the array entry
	text     = e[0]
	shortcut = e[1]
	callback = e[2] ? method( e[2] ).to_proc : nil
	data     = e[3]
	flags    = e[4]
	
	text2 = text
	if prefix_stack.size != 0
	  text2 = prefix_stack.join("/")+"/"+text2
	end
	@menubar.add( text2, 
		     shortcut, callback, data, flags )

	# if the flags have the SUBMENU bit set,
	# then the next items will be below this menu,
	# so we have to remember the name
	if (flags & Fltk::SUBMENU)==Fltk::SUBMENU
	  prefix_stack.push text
	end
      end
    }
  end # setup_menubar
  private :setup_menubar


  ##################################################################
  # The callbacks
  # These could have been implemented as blocks directly in the 
  # initialize method, but this way is more similar to the C++
  # way, and also cleaner when there starts getting code inside them.

  #################################
  # File callbacks

  ##
  # This callback function will clear the input widget and current filename.
  # It also calls the check_save() function to give the user the 
  # opportunity to save the current file first as needed. 
  def new_cb( widget, data )
    if @changed
      if !check_save then return end
    end
    @filename = ""
    @input.value=""
    set_changed(nil)
  end


  ##
  # This callback function will ask the user for a filename and then 
  # load the specified file into the input widget and current filename. 
  # It also calls the check_save() function to give the user the 
  # opportunity to save the current file first as needed:.
  def open_cb( widget, data )
    if @changed
      if !check_save then return end
    end
    newfile = Fltk::file_chooser("Open file?", "*.txt",@filename)
    if newfile
      load_file( newfile )
    end
  end

  ##
  # This callback saves the current file.
  # If the current filename is blank it calls the "save as" callback
  def save_cb( widget, data )
    if @filename == ""
      saveas_cb( widget, data )
    else
      save_file( @filename )
    end
  end

  ##
  # This callback asks the user for a filename and saves the current file.
  def saveas_cb( widget, data )
    newfile = Fltk::file_chooser("Save File As?", "*.txt",@filename)
    if newfile
      save_file( newfile )
    end
  end

  ##
  # The quit callback will first see if the current file has been modified,
  # and if so give the user a chance to save it. 
  # It then hides the main window.
  def quit_cb( widget, data )
    if @changed 
      if !check_save then return end
    end
    @window.hide
  end


  #################################
  # Edit callbacks

  def undo_cb( widget, data )
    @input.undo
  end

  def cut_cb( widget, data )
    @input.copy
    @input.cut
  end

  def copy_cb( widget, data )
    @input.copy
  end

  def paste_cb( widget, data )
    Fltk.paste( @input )    
  end

  def delete_cb( widget, data )
    @input.cut
  end

  #################################
  # Search callbacks

  ##
  # This callback function asks for a search string using 
  # the fl_input() convenience function and then calls 
  # the find2_cb() function to find the string. 
  def find_cb( widget, data )
    val = Fltk.input("Search String:", @search )
    if val then
      @search = val
      find2_cb( widget, data )
    end
  end

  ##
  # This function will find the next occurrence of the search string.
  # If the search string is blank then we want to pop up the search dialog.
  def find2_cb( widget, data )
    if ! @search then
      find_cb( widget, data )
      return
    end
    found = @input.value.index( @search, @input.mark )
    if found
      @input.position( found, found+@search.size )
    else
      Fltk.alert("No occurences of \'%s\' found!", @search )
    end      
  end

  ##
  # Shows the Replace dialog box.
  def replace_cb( widget, data )
    @replace_dlg.show
  end

  ##
  # Replaces the next occurance of the search string,
  # starting from the current position.
  def replace2_cb( widget, data )   
    find = @replace_find.value
    if ! find || find == "" then
      @replace_dlg.show
      return
    end
    found = @input.value.index( find, @input.mark )
    if found
      replacement = @replace_with.value
      @input.replace( found, found+find.size, 
		     replacement )
      @input.position( found + replacement.size )
    else
      Fltk.alert("No occurences of \'%s\' found!", find )
    end
  end

  ##
  # This callback will replace all occurences of the search
  # string in the file
  def replall_cb( widget, data )
    find = @replace_find.value
    if ! find || find == "" then
      @replace_dlg.show
      return
    end
    @input.position( 0 )
    times = 0
    found = true
    while found do
      found = @input.value.index( find, @input.mark )      
      if found
	replacement = @replace_with.value
	@input.replace( found, found+find.size, 
		       replacement )
	@input.position( found + replacement.size )
	times += 1
      end
    end
    if times > 0 then
      Fltk.message( "Replaced %d occurrences.", times )
    else
      Fltk.alert("No occurences of \'%s\' found!", find )
    end
  end

  ##
  # Replace dialog cancel callback.
  def replcan_cb( widget, data )
    @replace_dlg.hide
  end

  #################################
  # Other callbacks

  ##
  # Callback for noticing when the text has changed
  def changed_cb( widget, data )
    set_changed( true )
  end

  ##
  # sets the state wheter the text has been changed or not.
  # Adds the text " (modified)" next to the filename in the window
  # title if it is modified since last save or load.
  def set_changed( i )
    if i != @changed
      @changed = i
      title = if @filename==""
		"Untitled"
	      else
		@filename
	      end
      if @changed
	title +=" (modified)"
      end
      @window.label( title )
    end
  end

  ##
  # Check to see if we should save the file
  def check_save
    if ! @changed
      return true
    end    
    if Fltk.ask("The current file has not been saved.\n" +
		"Would you like to save it now?") then
      save_cb(nil,nil)
      return !@changed
    else
      return true
    end    
  end
  
  ##
  # Loads the file with the given name
  def load_file( file )
    begin
      File.open( file, "r"){|infile|
	s = infile.readlines.join
	@input.value = s
      }
      @input.position(0)
      @filename = file
      set_changed( false )

    rescue
      Fltk.alert("Unable to open \'%s\' for reading!", file)
    end
  end
  
  ##
  # Saves the text to the given file
  def save_file( file )
    begin
      File.open( file, "w"){|out|
	out << @input.value
      }
      @filename = file
      set_changed( false )
    rescue
      Fltk.alert("Unable to create \'%s\' for writing!",file)
    end
  end


end


# If this file is run from the command line,
# make an Editor object and start the Fltk event loop.
if __FILE__ == $0 then
  editor = Editor.new
  if( ARGV[0] )
    editor.load_file(ARGV[0])
  end
  Fltk.run
end
